`include "DecodeMacros.v"

module APMIPS(
    Clock,
    Enable,
    Reset,

    FetchAddress,
    FetchEnable,
    FetchIn,

    DataAddress,
    DataIn,
    DataInEnable,
    DataOut,
    DataOutEnable
);

parameter PCStartAddress = 32'h0040_0000;

`include "ControlConstants.v"

input                           Clock;
input                           Enable;
input                           Reset;

output      [31:0]              FetchAddress;
output                          FetchEnable;
input       [31:0]              FetchIn;

output      [31:0]              DataAddress;
input       [31:0]              DataIn;
output                          DataInEnable;
output      [31:0]              DataOut;
output                          DataOutEnable;


// Pipeline registers
reg         [31:0]              pc_from_pifif;

reg         [31:0]              pc_from_ifid;
reg                             branchtaken_from_ifid;
reg         [31:0]              instruction_from_ifid;

reg         [31:0]              pc_from_idex;
reg                             branchtaken_from_idex;
reg         [31:0]              instruction_from_idex;
reg         [31:0]              rs_from_idex;
reg         [31:0]              signextimm_from_idex;
reg         [31:0]              zeroextimm_from_idex;
reg                             operand0select_from_idex;
reg         [1:0]               operand1select_from_idex;
reg         [1:0]               pcselect_from_idex;
reg         [4:0]               aluopcode_from_idex;
reg         [25:0]              address_from_idex;
reg         [4:0]               shamt_from_idex;
reg         [4:0]               rsaddress_from_idex;
reg         [4:0]               rtaddress_from_idex;
reg         [31:0]              rt_from_idex;
reg         [4:0]               writebackaddress_from_idex;
reg                             memoryread_from_idex;
reg                             memorywrite_from_idex;

reg         [31:0]              instruction_from_expdf;
reg         [31:0]              rt_from_expdf;
reg         [4:0]               writebackaddress_from_expdf;
reg                             memoryread_from_expdf;
reg                             memorywrite_from_expdf;
reg         [31:0]              aluresult_from_expdf;
reg         [31:0]              nextpc_from_expdf;
reg                             flush_from_expdf;

reg         [31:0]              instruction_from_pdfdf;
reg         [4:0]               writebackaddress_from_pdfdf;
reg                             memoryread_from_pdfdf;
reg         [31:0]              aluresult_from_pdfdf;

reg         [31:0]              instruction_from_dfwb;
reg         [4:0]               writebackaddress_from_dfwb;
reg                             memoryread_from_dfwb;
reg         [31:0]              aluresult_from_dfwb;
reg         [31:0]              readdata_from_dfwb;


reg                             reset;

wire                            flush_from_hdu;
wire                            halfstall_from_hdu;

wire                            fwdpdfrs_from_fdu;
wire                            fwdpdfrt_from_fdu;
wire                            fwddfrs_from_fdu;
wire                            fwddfrt_from_fdu;
wire                            fwdwbrs_from_fdu;
wire                            fwdwbrt_from_fdu;

reg         [31:0]              nextpc_from_ex;

wire        [31:0]              rs_from_id;
wire        [31:0]              rt_from_id;
wire        [15:0]              immediate_from_id;
wire        [31:0]              signextimm_from_id;
wire        [31:0]              zeroextimm_from_id;
wire        [4:0]               shamt_from_id;
wire        [4:0]               rsaddress_from_id;
wire        [4:0]               rtaddress_from_id;
wire        [4:0]               rdaddress_from_id;
wire        [25:0]              address_from_id;
reg         [4:0]               writebackaddress_from_id;

reg         [31:0]              rs_from_ex;
reg         [31:0]              rt_from_ex;

reg         [31:0]              operand0_from_ex;
reg         [31:0]              operand1_from_ex;

wire        [31:0]              pcplus4_from_ex;
wire        [31:0]              branchaddress_from_ex;
wire        [31:0]              jumpaddress_from_ex;
wire        [31:0]              nextpcplus4_from_ex;

wire        [31:0]              aluresult_from_ex;
wire                            branch_from_ex;

wire        [31:0]              writebackdata_from_wb;

wire                            operand0select_from_ctrl;
wire        [1:0]               operand1select_from_ctrl;
wire        [4:0]               aluopcode_from_ctrl;
wire        [1:0]               pcselect_from_ctrl;
                                
wire                            memoryread_from_ctrl;
wire                            memorywrite_from_ctrl;

wire        [1:0]               writebackselect_from_ctrl;

always@(posedge Clock) begin
    if(Enable) begin
        if(Reset) begin
            pc_from_pifif               <= PCStartAddress + 4;

            pc_from_ifid                <= PCStartAddress + 8;
            branchtaken_from_ifid       <= 1'b0;
            instruction_from_ifid       <= 32'h0000_0000;
            
            pc_from_idex                <= PCStartAddress + 12;
            branchtaken_from_idex       <= 1'b0;
            instruction_from_idex       <= 32'h0000_0000;
            rs_from_idex                <= 32'h0000_0000;
            signextimm_from_idex        <= 32'h0000_0000;
            zeroextimm_from_idex        <= 32'h0000_0000;
            operand0select_from_idex    <= 1'b0;
            operand1select_from_idex    <= 2'b00;
            pcselect_from_idex          <= 2'b00;
            aluopcode_from_idex         <= 5'b00000;
            address_from_idex           <= 5'b00000;
            shamt_from_idex             <= 5'b00000;
            rsaddress_from_idex         <= 5'b00000;
            rtaddress_from_idex         <= 5'b00000;
            rt_from_idex                <= 32'h0000_0000;
            writebackaddress_from_idex  <= 5'b00000;
            memoryread_from_idex        <= 1'b0;
            memorywrite_from_idex       <= 1'b0;

            instruction_from_expdf      <= 32'h0000_0000;
            rt_from_expdf               <= 32'h0000_0000;
            writebackaddress_from_expdf <= 5'b00000;
            memoryread_from_expdf       <= 1'b0;
            memorywrite_from_expdf      <= 1'b0;
            aluresult_from_expdf        <= 32'h0000_0000;
            nextpc_from_expdf           <= PCStartAddress;
            flush_from_expdf            <= 1'b0;

            instruction_from_pdfdf      <= 32'h0000_0000;
            writebackaddress_from_pdfdf <= 5'b00000;
            memoryread_from_pdfdf       <= 1'b0;
            aluresult_from_pdfdf        <= 32'h0000_0000;
            
            instruction_from_dfwb       <= 32'h0000_0000;
            writebackaddress_from_dfwb  <= 5'b00000;
            memoryread_from_dfwb        <= 1'b0;
            aluresult_from_dfwb         <= 32'h0000_0000;
            readdata_from_dfwb          <= 32'h0000_0000;

            reset                       <= 1'b1;
        end else if(flush_from_expdf) begin
            pc_from_pifif               <= nextpc_from_expdf;

            pc_from_ifid                <= pc_from_pifif;
            branchtaken_from_ifid       <= 1'b0;
            instruction_from_ifid       <= 32'h0000_0000;
            
            pc_from_idex                <= pc_from_ifid;
            branchtaken_from_idex       <= 1'b0;
            instruction_from_idex       <= 32'h0000_0000;
            rs_from_idex                <= 32'h0000_0000;
            signextimm_from_idex        <= 32'h0000_0000;
            zeroextimm_from_idex        <= 32'h0000_0000;
            operand0select_from_idex    <= 1'b0;
            operand1select_from_idex    <= 2'b00;
            pcselect_from_idex          <= 2'b00;
            aluopcode_from_idex         <= 5'b00000;
            address_from_idex           <= 5'b00000;
            shamt_from_idex             <= 5'b00000;
            rsaddress_from_idex         <= 5'b00000;
            rtaddress_from_idex         <= 5'b00000;
            rt_from_idex                <= 32'h0000_0000;
            writebackaddress_from_idex  <= 5'b00000;
            memoryread_from_idex        <= 1'b0;
            memorywrite_from_idex       <= 1'b0;

            instruction_from_expdf      <= instruction_from_idex;
            rt_from_expdf               <= rt_from_ex;
            writebackaddress_from_expdf <= writebackaddress_from_idex;
            memoryread_from_expdf       <= memoryread_from_idex;
            memorywrite_from_expdf      <= memorywrite_from_idex;
            aluresult_from_expdf        <= aluresult_from_ex;
            nextpc_from_expdf           <= nextpc_from_ex;
            flush_from_expdf            <= flush_from_hdu;

            instruction_from_pdfdf      <= instruction_from_expdf;
            writebackaddress_from_pdfdf <= writebackaddress_from_expdf;
            memoryread_from_pdfdf       <= memoryread_from_expdf;
            aluresult_from_pdfdf        <= aluresult_from_expdf;
            
            instruction_from_dfwb       <= instruction_from_pdfdf;
            writebackaddress_from_dfwb  <= writebackaddress_from_pdfdf;
            memoryread_from_dfwb        <= memoryread_from_pdfdf;
            aluresult_from_dfwb         <= aluresult_from_pdfdf;
            readdata_from_dfwb          <= DataIn;

            reset                       <= 1'b0;
        end else if(halfstall_from_hdu) begin
            rs_from_idex                <= rs_from_ex;
            rt_from_idex                <= rt_from_ex;

            instruction_from_expdf      <= 32'h0000_0000;
            rt_from_expdf               <= 32'h0000_0000;
            writebackaddress_from_expdf <= 5'b00000;
            memoryread_from_expdf       <= 1'b0;
            memorywrite_from_expdf      <= 1'b0;
            aluresult_from_expdf        <= 32'h0000_0000;
            nextpc_from_expdf           <= nextpc_from_expdf;
            flush_from_expdf            <= 1'b0;

            instruction_from_pdfdf      <= instruction_from_expdf;
            writebackaddress_from_pdfdf <= writebackaddress_from_expdf;
            memoryread_from_pdfdf       <= memoryread_from_expdf;
            aluresult_from_pdfdf        <= aluresult_from_expdf;
            
            instruction_from_dfwb       <= instruction_from_pdfdf;
            writebackaddress_from_dfwb  <= writebackaddress_from_pdfdf;
            memoryread_from_dfwb        <= memoryread_from_pdfdf;
            aluresult_from_dfwb         <= aluresult_from_pdfdf;
            readdata_from_dfwb          <= DataIn;

            reset                       <= 1'b0;
        end else begin
            pc_from_pifif               <= nextpc_from_expdf;

            pc_from_ifid                <= pc_from_pifif;
            branchtaken_from_ifid       <= 1'b0;
            instruction_from_ifid       <= reset ? 32'h0000_0000 : FetchIn;
            
            pc_from_idex                <= pc_from_ifid;
            branchtaken_from_idex       <= branchtaken_from_ifid;
            instruction_from_idex       <= instruction_from_ifid;
            rs_from_idex                <= rs_from_id;
            signextimm_from_idex        <= signextimm_from_id;
            zeroextimm_from_idex        <= zeroextimm_from_id;
            operand0select_from_idex    <= operand0select_from_ctrl;
            operand1select_from_idex    <= operand1select_from_ctrl;
            pcselect_from_idex          <= pcselect_from_ctrl;
            aluopcode_from_idex         <= aluopcode_from_ctrl;
            address_from_idex           <= address_from_id;
            shamt_from_idex             <= shamt_from_id;
            rsaddress_from_idex         <= rsaddress_from_id;
            rtaddress_from_idex         <= rtaddress_from_id;
            rt_from_idex                <= rt_from_id;
            writebackaddress_from_idex  <= writebackaddress_from_id;
            memoryread_from_idex        <= memoryread_from_ctrl;
            memorywrite_from_idex       <= memorywrite_from_ctrl;

            instruction_from_expdf      <= instruction_from_idex;
            rt_from_expdf               <= rt_from_ex;
            writebackaddress_from_expdf <= writebackaddress_from_idex;
            memoryread_from_expdf       <= memoryread_from_idex;
            memorywrite_from_expdf      <= memorywrite_from_idex;
            aluresult_from_expdf        <= aluresult_from_ex;
            nextpc_from_expdf           <= nextpc_from_ex;
            flush_from_expdf            <= flush_from_hdu;

            instruction_from_pdfdf      <= instruction_from_expdf;
            writebackaddress_from_pdfdf <= writebackaddress_from_expdf;
            memoryread_from_pdfdf       <= memoryread_from_expdf;
            aluresult_from_pdfdf        <= aluresult_from_expdf;
            
            instruction_from_dfwb       <= instruction_from_pdfdf;
            writebackaddress_from_dfwb  <= writebackaddress_from_pdfdf;
            memoryread_from_dfwb        <= memoryread_from_pdfdf;
            aluresult_from_dfwb         <= aluresult_from_pdfdf;
            readdata_from_dfwb          <= DataIn;

            reset                       <= 1'b0;
        end
    end
end

// PIF Stage
assign FetchAddress = nextpc_from_expdf;
assign FetchEnable  = Enable && !halfstall_from_hdu;

// Hazard Detection Unit
assign flush_from_hdu = (pcselect_from_idex == PCSELBRANCH && branch_from_ex != branchtaken_from_idex) ||
                        (pcselect_from_idex == PCSELJMPIMM) || (pcselect_from_idex == PCSELJMPREG);
assign halfstall_from_hdu = ((fwdpdfrs_from_fdu || fwdpdfrt_from_fdu) && memoryread_from_expdf) ||
                            ((fwddfrs_from_fdu || fwddfrt_from_fdu) && memoryread_from_pdfdf);


// ID Stage
assign immediate_from_id    = `IMMEDIATE(instruction_from_ifid);
assign signextimm_from_id   = {{16{immediate_from_id[15]}}, immediate_from_id};
assign zeroextimm_from_id   = {16'b0, immediate_from_id};
assign shamt_from_id        = `SHAMT(instruction_from_ifid);
assign rsaddress_from_id    = `RS(instruction_from_ifid);
assign rtaddress_from_id    = `RT(instruction_from_ifid);
assign rdaddress_from_id    = `RD(instruction_from_ifid);
assign address_from_id      = `ADDRESS(instruction_from_ifid);

always@(*) begin
    case(writebackselect_from_ctrl)
        WBSELNONE: writebackaddress_from_id = 5'b0;
        WBSELRT:   writebackaddress_from_id = rtaddress_from_id;
        WBSELRD:   writebackaddress_from_id = rdaddress_from_id;
        WBSEL31:   writebackaddress_from_id = 5'd31;
        default:   writebackaddress_from_id = 5'bx;
    endcase
end

RegisterFile regfile(
    .Clock(                     Clock),
    .Enable(                    Enable),

    .ReadAddress0(              rsaddress_from_id),
    .ReadData0(                 rs_from_id),
    .ReadAddress1(              rtaddress_from_id),
    .ReadData1(                 rt_from_id),

    .WriteAddress(              writebackaddress_from_dfwb),
    .WriteEnable(               1'b1),
    .WriteData(                 writebackdata_from_wb)
);

Control control(
    .Instruction(               instruction_from_ifid),

    .Operand0Select(            operand0select_from_ctrl),
    .Operand1Select(            operand1select_from_ctrl),
    .ALUOpcode(                 aluopcode_from_ctrl),
    .PCSelect(                  pcselect_from_ctrl),
    .MemoryRead(                memoryread_from_ctrl),
    .MemoryWrite(               memorywrite_from_ctrl),
    .WriteBackSelect(           writebackselect_from_ctrl)
);


// EX Stage
assign pcplus4_from_ex       = pc_from_idex + 4;
assign branchaddress_from_ex = pcplus4_from_ex + {signextimm_from_idex, 2'b00};
assign jumpaddress_from_ex   = {pcplus4_from_ex[31:28], address_from_idex, 2'b00};
assign nextpcplus4_from_ex   = nextpc_from_expdf + 4;

always@(*) begin
    case(pcselect_from_idex)
        PCSELPLUS4:   nextpc_from_ex = nextpcplus4_from_ex;
        PCSELBRANCH:  nextpc_from_ex = branch_from_ex ? branchaddress_from_ex : nextpcplus4_from_ex;
        PCSELJMPIMM:  nextpc_from_ex = jumpaddress_from_ex;
        PCSELJMPREG:  nextpc_from_ex = rs_from_ex;
        default:      nextpc_from_ex = 32'bx;
    endcase
end

always@(*) begin
    if(fwdpdfrs_from_fdu) begin
        rs_from_ex = aluresult_from_expdf;
    end else if(fwddfrs_from_fdu) begin
        rs_from_ex = aluresult_from_pdfdf;
    end else if(fwdwbrs_from_fdu) begin
        rs_from_ex = writebackdata_from_wb;
    end else begin
        rs_from_ex = rs_from_idex;
    end
    if(fwdpdfrt_from_fdu) begin
        rt_from_ex = aluresult_from_expdf;
    end else if(fwddfrt_from_fdu) begin
        rt_from_ex = aluresult_from_pdfdf;
    end else if(fwdwbrt_from_fdu) begin
        rt_from_ex = writebackdata_from_wb;
    end else begin
        rt_from_ex = rt_from_idex;
    end
    case(operand0select_from_idex)
        OP0SELRS:    operand0_from_ex = rs_from_ex;
        OP0SELSHAMT: operand0_from_ex = shamt_from_idex;
        default:     operand0_from_ex = 32'bx;
    endcase
    case(operand1select_from_idex)
        OP1SELRT:    operand1_from_ex = rt_from_ex;
        OP1SELSEIMM: operand1_from_ex = signextimm_from_idex;
        OP1SELZEIMM: operand1_from_ex = zeroextimm_from_idex;
        OP1SELPC:    operand1_from_ex = pc_from_idex + 8;
        default:     operand1_from_ex = 32'bx;
    endcase
end

ALU alu(
    .Operand0(                  operand0_from_ex),
    .Operand1(                  operand1_from_ex),
    .Opcode(                    aluopcode_from_idex),

    .Result(                    aluresult_from_ex),
    .Overflow(                  ),
    .Branch(                    branch_from_ex)
);

// Forwarding Unit
assign fwdpdfrs_from_fdu = writebackaddress_from_expdf != 5'b0 && writebackaddress_from_expdf == rsaddress_from_idex;
assign fwdpdfrt_from_fdu = writebackaddress_from_expdf != 5'b0 && writebackaddress_from_expdf == rtaddress_from_idex;

assign fwddfrs_from_fdu = writebackaddress_from_pdfdf != 5'b0 && writebackaddress_from_pdfdf == rsaddress_from_idex;
assign fwddfrt_from_fdu = writebackaddress_from_pdfdf != 5'b0 && writebackaddress_from_pdfdf == rtaddress_from_idex;

assign fwdwbrs_from_fdu = writebackaddress_from_dfwb != 5'b0 && writebackaddress_from_dfwb == rsaddress_from_idex;
assign fwdwbrt_from_fdu = writebackaddress_from_dfwb != 5'b0 && writebackaddress_from_dfwb == rtaddress_from_idex;


// PDF Stage
assign DataAddress   = aluresult_from_expdf;
assign DataInEnable  = memoryread_from_expdf;
assign DataOutEnable = memorywrite_from_expdf;
assign DataOut       = rt_from_expdf;


// WB Stage
assign writebackdata_from_wb = memoryread_from_dfwb ? readdata_from_dfwb : aluresult_from_dfwb;


endmodule
