module InstructionMemory(
    Clock,
    Enable,

    FetchAddress,
    FetchOut,
    FetchEnable,

    DataAddress,
    DataIn,
    DataInEnable,
    DataOut,
    DataOutEnable
);

parameter DataWidth = 32,
          AddressWidth = 12,
          InitialHexFile = "";

localparam Depth = 2 ** (AddressWidth-2);


input                           Clock;
input                           Enable;

input       [AddressWidth-1:0]  FetchAddress;
output      [DataWidth-1:0]     FetchOut;
input                           FetchEnable;

input       [AddressWidth-1:0]  DataAddress;
input       [DataWidth-1:0]     DataIn;
input                           DataInEnable;
output      [DataWidth-1:0]     DataOut;
input                           DataOutEnable;

reg         [AddressWidth-3:0]  fetchaddress;

reg         [AddressWidth-3:0]  dataaddress;

reg         [DataWidth-1:0]     state[Depth-1:0];


generate if(InitialHexFile != "")
initial
    $readmemh(InitialHexFile, state);
endgenerate

assign FetchOut = state[fetchaddress];
assign DataOut = state[dataaddress];

always@(posedge Clock) begin
    if(Enable) begin
        if(FetchEnable) begin
            fetchaddress <= FetchAddress[AddressWidth-1:2];
        end

        if(DataOutEnable) begin
            dataaddress <= DataAddress[AddressWidth-1:2];
        end

        if(DataInEnable) begin
            state[DataAddress[AddressWidth-1:2]] <= DataIn;
        end
    end
end

endmodule
