`timescale 1ns / 1ps

module ML505Top(
    GPIO_SW_C,
    USER_CLK,

    FPGA_SERIAL_RX0,
    FPGA_SERIAL_TX0,

    FPGA_SERIAL_RX1,
    FPGA_SERIAL_TX1
);

input                           GPIO_SW_C;
input                           USER_CLK;

input                           FPGA_SERIAL_RX0;
output                          FPGA_SERIAL_TX0;

input                           FPGA_SERIAL_RX1;
output                          FPGA_SERIAL_TX1;

// CPU
wire        [31:0]              fetchaddress_from_cpu;
wire                            fetchenable_from_cpu;

wire        [31:0]              dataaddress_from_cpu;
wire                            datainenable_from_cpu;
wire        [31:0]              dataout_from_cpu;
wire                            dataoutenable_from_cpu;

// Instruction Memory
wire        [31:0]              fetchout_from_imem;

// Data Memory

ClockSource125MHz clocksource125(
    .Clock100MHz(               USER_CLK),
    .Clock125MHz(               clock125mhz)
);

ResetControl resetcontrol(
    .Clock(                     clock125mhz),

    .CPUTrigger(                GPIO_SW_C),
    .CPUReset(                  cpureset)
);

APMIPS apmips(
    .Clock(                     clock125mhz),
    .Enable(                    1'b1),
    .Reset(                     cpureset),

    .FetchAddress(              fetchaddress_from_cpu),
    .FetchEnable(               fetchenable_from_cpu),
    .FetchIn(                   fetchout_from_imem),

    .DataAddress(               dataaddress_from_cpu),
    .DataIn(                    ),
    .DataInEnable(              datainenable_from_cpu),
    .DataOut(                   dataout_from_cpu),
    .DataOutEnable(             dataoutenable_from_cpu)
);

// Instruction Memory
InstructionMemory #(
    .DataWidth(                 32),
    .AddressWidth(              16),
    .InitialHexFile(            "")
) imem(
    .Clock(                     clock125mhz),
    .Enable(                    1'b1),

    .FetchAddress(              fetchaddress_from_cpu),
    .FetchOut(                  fetchout_from_imem),
    .FetchEnable(               fetchenable_from_cpu),

    .DataAddress(               ),
    .DataIn(                    ),
    .DataInEnable(              ),
    .DataOut(                   ),
    .DataOutEnable(             )
);

// Data Memory
DataMemory #(
    .DataWidth(                 32),
    .AddressWidth(              16),
    .InitialHexFile(            "")
) dmem(
    .Clock(                     clock125mhz),
    .Enable(                    1'b1),

    .DataAddress(               ),
    .DataIn(                    ),
    .DataInEnable(              ),
    .DataOut(                   ),
    .DataOutEnable(             )
);

endmodule
